"""Command-line interface for creating and adminitrating template projects

Defines module functions for sub-commands
"""
import sys
import platform
import logging
import argparse
import devopstemplate
from devopstemplate.config import CommandsConfig, ProjectConfig
from devopstemplate.template import DevOpsTemplate


def create(args):
    """Wrapper for sub-command create

    Params:
        args: argparse.Namespace object with argument parser attributes
    """

    config = ProjectConfig(args)
    template = DevOpsTemplate(projectdirectory=config.project_dir,
                              overwrite_exists=config.overwrite_exists,
                              skip_exists=config.skip_exists,
                              dry_run=config.dry_run)

    param_dict, comp_list = config.create()
    template.create(context=param_dict, components=comp_list)


def manage(args):
    """Wrapper for sub-command manage

    Params:
        args: argparse.Namespace object with argument parser attributes
    """
    config = ProjectConfig(args)
    template = DevOpsTemplate(projectdirectory=config.project_dir,
                              overwrite_exists=config.overwrite_exists,
                              skip_exists=config.skip_exists,
                              dry_run=config.dry_run)

    param_dict, comp_list = config.manage()
    template.manage(context=param_dict, components=comp_list)


def cookiecutter(args):
    """Wrapper for sub-command manage

    Params:
        args: argparse.Namespace object with argument parser attributes
    """
    config = ProjectConfig(args)
    template = DevOpsTemplate(projectdirectory=config.project_dir,
                              overwrite_exists=config.overwrite_exists,
                              skip_exists=config.skip_exists,
                              dry_run=config.dry_run)

    param_dict, comp_list = config.cookiecutter()
    template.cookiecutter(context=param_dict, components=comp_list)


def arg_command_group(parser, group_name, group_argument_list):
    """Add a group of optional arguments to the parser.

    Params:
        parser: argparse.ArgumentParser where the argument group will be added.
        group_name: String with the name of the argument group.
        group_argument_list: List of dict objects where each dict specifies an
            argument.
    Returns:
        group: The argument group object that has been created for the parser.
    Raises:
        ValueError: if the group_argument_list is empty
    """
    if not group_argument_list:
        raise ValueError('Invalid group_argument_list')

    # Add argument group
    group = parser.add_argument_group(group_name)

    # Add arguments
    for arg_dict in group_argument_list:
        arg_name = arg_dict['name']
        arg_name = f'--{arg_name}'
        arg_help = arg_dict['help']
        arg_value = arg_dict['default']
        if isinstance(arg_value, bool):
            # Attention: always interpret boolean flag in a positive sense
            # The arg_value specifies where to add the corresponding template
            # component by default. The presence of a boolean flag negates the
            # corresponding boolean action.
            group.add_argument(arg_name, action='store_true', help=arg_help)
        else:
            group.add_argument(arg_name, default=arg_value, help=arg_help)

    return group


def parse_args(args_list):
    """Parse command-line arguments and call a function for processing user
    request.

    The parser defines (sub) commands which are processed with the 'func'
    attribute function obtained from the parsing result, i.e., Namespace object
    ( Namespace=parser.parse_args(args) ).
    Each command sets its function to the func attribute with set_defaults.
    After all command-line flags have been processed, the function associated
    with func is executed.

    Params:
        args_list: List of strings with command-line flags (sys.argv[1:])
    """
    logger = logging.getLogger('main.parse_args')

    # Initiate CommandsConfig in order to obtain command definitions
    cfg = CommandsConfig()

    descr = ''.join(['Create and manage dev-ops template projects. '])
    parser = argparse.ArgumentParser(description=descr)
    # top-level arguments (optional)
    parser.add_argument('--project-dir', default='.',
                        help='Project directory, default: current directory')
    parser.add_argument('--skip-exists', action='store_true',
                        help=('Skip copying files if they already '
                              'exist in the project directory'))
    parser.add_argument('--overwrite-exists', action='store_true',
                        help=('Overwrite files if they already '
                              'exist in the project directory'))
    parser.add_argument('--quiet', action='store_true',
                        help='Print only warning/error messages')
    parser.add_argument('--verbose', action='store_true',
                        help='Print debug messages')
    parser.add_argument('--dry-run', action='store_true',
                        help='Pretend to perform actions')
    parser.add_argument('--version', action='store_true',
                        help='Print version')
    # Default for printing help message if no command is provided
    # attribute 'func' is set to a lambda function
    # --> if attribute func keeps the lambda function until the entire parser
    # has been evaluated: call print_help() and show the help message
    parser.set_defaults(func=lambda _: parser.print_help())
    # Subparser commands for project creation and management
    subparsers = parser.add_subparsers(help='Commands')

    create_parser = subparsers.add_parser('create',
                                          help=('Create a new project based '
                                                'on the dev-ops template'))
    create_parser.add_argument('-i', '--interactive', action='store_true',
                               help=('Configure project parameters/components'
                                     ' interactively'))

    arg_command_group(create_parser, 'project parameters',
                      group_argument_list=cfg.values('create', 'parameters'))
    arg_command_group(create_parser, 'project components',
                      group_argument_list=cfg.values('create', 'components'))
    # If the create subparser has been activated by the 'create' command,
    # override the func attribute with a pointer to the 'create' function
    # (defined above) --> overrides default defined for the main parser.
    create_parser.set_defaults(func=create)

    manage_parser = subparsers.add_parser('manage',
                                          help=('Add individual components of '
                                                'the dev-ops template'))
    arg_command_group(manage_parser, 'project components',
                      group_argument_list=cfg.values('manage', 'components'))
    # If the manage subparser has been activated by the 'manage' command,
    # override the func attribute with a pointer to the 'manage' function
    # (defined above) --> overrides default defined for the main parser.
    manage_parser.set_defaults(func=manage)

    cc_parser = subparsers.add_parser('cookiecutter',
                                      help=('Create a cookiecutter'
                                            ' template'))
    cc_parser.add_argument('-i', '--interactive', action='store_true',
                           help=('Configure project parameters/components'
                                 ' interactively'))
    arg_command_group(cc_parser, 'project parameters',
                      group_argument_list=cfg.values('cookiecutter',
                                                     'parameters'))
    arg_command_group(cc_parser, 'project components',
                      group_argument_list=cfg.values('cookiecutter',
                                                     'components'))
    # If the cookiecutter subparser has been activated by the 'cookiecutter'
    # command, override the func attribute with a pointer to the 'cookiecutter'
    # function (defined above) --> overrides default for the main parser.
    cc_parser.set_defaults(func=cookiecutter)
    args_ns = parser.parse_args(args=args_list)

    # If version flag is set: print version and quit
    if args_ns.version:
        logger.info('devopstemplate v%s', devopstemplate.__version__)
        return

    # Set log level according to command-line flags
    if args_ns.verbose:
        devopstemplate.LOGCONFIG.debug()
        logger.debug('%s:: %s', platform.node(), ' '.join(sys.argv))
        logger.debug('devopstemplate v%s\n', devopstemplate.__version__)
    elif args_ns.quiet:
        devopstemplate.LOGCONFIG.warning()

    logger.debug('Command-line args: %s', args_list)
    args_dict = vars(args_ns)
    logger.debug('Options:\n%s', ', '.join(f'{key} : {val}'
                                           for key, val in args_dict.items()))
    args_ns.func(args_ns)


def main():
    """Entrypoint for starting the command-line interface.
    Control-flow continues depending on user arguments.

    Passes all command-line flags sys.argv[1:] to argparse
    (implemented in parse_args)
    """
    parse_args(sys.argv[1:])


if __name__ == "__main__":
    main()



import pytest
from argparse import Namespace
import {{project_slug}}.main


@pytest.fixture
def args_ns_ref():
    return Namespace()


def test_parseargs(args_ns_ref):
    arg_list = []
    args_ns = {{project_slug}}.main.parse_args(arg_list)
    # Define reference/expected result
    args_ns_ref.verbose = False
    args_ns_ref.quiet = False
    assert args_ns == args_ns_ref


def test_parseargs_verbose(args_ns_ref):
    arg_list = ['--verbose']
    args_ns = {{project_slug}}.main.parse_args(arg_list)
    args_ns_ref.verbose = True
    args_ns_ref.quiet = False
    assert args_ns == args_ns_ref


def test_parseargs_quiet(args_ns_ref):
    arg_list = ['--quiet']
    args_ns = {{project_slug}}.main.parse_args(arg_list)
    args_ns_ref.verbose = False
    args_ns_ref.quiet = True
    assert args_ns == args_ns_ref


if __name__ == "__main__":
    pytest.main()



import logging
import datetime
from pprint import pformat
from pymongo import MongoClient


def mongo_sample():
    host = '127.0.0.1'
    client = MongoClient(host=host)
    db = client.test_database
    list_db_collections(db)
    results_collection = db.results
    res = {'model': 'model_id_1',
           'dataset': 'dataset_id_1',
           'result': 'X% acc',
           'timestamp': str(datetime.datetime.now())
           }
    results_collection.insert_one(res)
    list_db_collections(db)


def list_db_collections(db):
    logger = logging.getLogger('list_db_collections')
    collection_list = db.list_collection_names()
    logger.info('collections:\n%s', ', '.join(collection_list))
    for col_id in collection_list:
        collection = db[col_id]
        logger.info('collection: %s', col_id)
        for entry in collection.find():
            logger.info('\n%s', pformat(entry))
        logger.info('-')


if __name__ == "__main__":
    LOGGING_FORMAT = '%(asctime)-15s: [%(name)s] %(message)s'
    # LOGGING_FORMAT = '[%(name)s] %(message)s'
    logging.basicConfig(level=logging.INFO,
                        format=LOGGING_FORMAT)
    mongo_sample()



"""Initialize the Python logging system"""
import logging


class LoggerConfig():
    """
    Useful log message formats:
    LOGGING_FORMAT = '%(asctime)-15s: [%(name)s] %(message)s'
    LOGGING_FORMAT = '[%(name)s] %(message)s'
    LOGGING_FORMAT = '%(message)s'
    """

    def __init__(self):
        """Initialize log config with default level info and a short log
        message (contains just the message and not meta information)
        """
        # Default log level: info
        # logging.basicConfig(level=logging.INFO,
        #                     format='%(message)s')
        self.__format_plain = '%(message)s'
        self.__format_debug = '%(asctime)-15s: [%(name)s] %(message)s'
        self.__handler = logging.StreamHandler()
        logging.getLogger().addHandler(self.__handler)
        self.info()

    def debug(self):
        """Switch to debug log level. Change level-of-detail for log message.
        (add meta information)
        """
        logging.getLogger().setLevel(logging.DEBUG)
        self.__handler.setFormatter(logging.Formatter(self.__format_debug))

    def info(self):
        """Switch to info log level. Change level-of-detail for log message.
        (just the log message)
        """
        logging.getLogger().setLevel(logging.INFO)
        self.__handler.setFormatter(logging.Formatter(self.__format_plain))

    def warning(self):
        """Switch to info log level. Change level-of-detail for log message.
        (just the log message)
        """
        logging.getLogger().setLevel(logging.WARNING)
        self.__handler.setFormatter(logging.Formatter(self.__format_plain))



"""Top-level import package for {{project_slug}}"""
from {{project_slug}}.log import LoggerConfig


# Package meta information
__version__ = '{{project_version}}'
# Logger configuration can be accessed through global variable LOGCONFIG
LOGCONFIG = LoggerConfig()



"""Initializations on Python package level

- Define package version
- Initialize Python logging framework
"""

from devopstemplate.log import LoggerConfig

# Version can be parsed from setup.py or managed globally,
# e.g., with bumpversion
__version__ = '0.9.0'


# Global module variable that stores a singelton of the LoggerConfig
# Can be used globally in order to adjust log behavior.
LOGCONFIG = LoggerConfig()



"""Use the quickstart example in order to test your MlFlow container:
https://mlflow.org/docs/latest/quickstart.html#using-the-tracking-api

Here, the example is extended by logging to an MlFlow tracking server
(instead of the local directory),
by storing artifacts to the (S3 compatible) MinIO object storage
and by using a single mlflow run with a given experiment name.
"""
import os
from random import random, randint
import botocore
import boto3
import mlflow


def create_bucket(bucket_name):
    s3cli = boto3.client('s3',
                         endpoint_url=os.environ.get('MLFLOW_S3_ENDPOINT_URL'))
    try:
        s3cli.create_bucket(Bucket=bucket_name)
        print(f'S3 bucket {bucket_name} has been created')
    except botocore.exceptions.ClientError as error:
        if (error.response['Error']['Code'] == 'BucketAlreadyOwnedByYou' or
                error.response['Error']['Code'] == 'BucketAlreadyExists'):
            print('Info: bucket already exists')
        else:
            raise error


def main():
    # S3 storage configurations are provided as environment variables which
    # is a requirement of the AWS boto3 library used in the mlflow client
    # S3 storage credentials as defined for the MinIO server
    os.environ['AWS_ACCESS_KEY_ID'] = 'minio'
    os.environ['AWS_SECRET_ACCESS_KEY'] = 'minio_secret'
    # MinIO server URL (S3 compatible)
    # (adjust host:port if run within Docker network)
    os.environ['MLFLOW_S3_ENDPOINT_URL'] = 'http://localhost:9001'
    # Create S3 bucket if it does not exist already
    # Name of the bucket as defined for the MlFlow server (ARTIFACT_ROOT)
    # e.g., s3://mlflow/
    bucket_name = 'mlflow'
    create_bucket(bucket_name)
    # MlFlow tracking server URL
    # (adjust host:port if run within Docker network)
    mlflow.set_tracking_uri('http://localhost:5000')
    # Create new experiment
    mlflow_exp_id = mlflow.set_experiment(experiment_name='random-test')
    # Start logging experiment data in new run for the given experiment
    with mlflow.start_run(experiment_id=mlflow_exp_id):
        # Log a parameter (key-value pair)
        mlflow.log_param("param1", randint(0, 100))

        # Log a metric; metrics can be updated throughout the run
        mlflow.log_metric("foo", random())
        mlflow.log_metric("foo", random() + 1)
        mlflow.log_metric("foo", random() + 2)

        # Log an artifact (output file)
        if not os.path.exists("outputs"):
            os.makedirs("outputs")
        with open("outputs/test.txt", "w") as f:
            f.write("hello world!")
        print(mlflow.get_artifact_uri())
        mlflow.log_artifacts("outputs")


if __name__ == "__main__":
    main()



"""Create new projects from template and administer existing projects
    - defines which components/files will be installed for user requests
    - defines how components/files will be installed according to user requests
"""
import os
import logging
import json
from jinja2 import Environment, PackageLoader, select_autoescape
from jinja2 import Template
from devopstemplate import pkg


class SkipFileError(FileExistsError):
    """Will be raised if a file that is to be created already exists and should
    be skipped according to user flags.
    """


class DevOpsTemplate():
    """Create and modify instances of the DevOps template:

    create: New instance of the template
    manage: Modify an existing instance of the template
    cookiecutter: generate cookiecutter template from the devops template
    """

    def __init__(self, projectdirectory,
                 overwrite_exists=False, skip_exists=False, dry_run=False):
        """Provide configurations that are common to all DevOpsTemplate actions

        Params:
            projectdirectory: String with a (relative) path to the directory
                that contains the instance of the template.
            overwrite_exists: Boolean specifying if existing files should be
                overwritten. An error is raised otherwise.
            skip_exists: Boolean specifying if existing files should be
                skipped/ignores. An error is raised otherwise.
            dry_run: Boolean specifying whether to not perform any actions in
                order to see (in the log) what would have happend.
        """
        self.__projectdir = projectdirectory
        self.__overwrite = overwrite_exists
        self.__skip = skip_exists
        self.__dry_run = dry_run
        with pkg.stream('template.json') as handle:
            self.__template_dict = json.load(handle)
        self.__template_dname = 'template'
        # ATTENTION: using __package__ may only work as long as this module
        # (template.py) is located in the top-level import directory
        loader = PackageLoader(__package__,
                               self.__template_dname)
        self.__env = Environment(loader=loader,
                                 autoescape=select_autoescape(default=True))
        # Create project base directory if not present
        self.__mkdir(projectdirectory)

    def create(self, context, components):
        """Create a new project from the DevOps template given config options.

        Installs components which are defined in template.json

        Params:
            context: Dictionary with configuration flags supported by the
                template (typically generated by the CLI, see main.create).
            components: Template components that should be installed.
        """
        logger = logging.getLogger('DevOpsTemplate.create')
        logger.info('Create project from template')
        logger.info('Project name: %s', context['project_name'])
        logger.info('Package name: %s', context['project_slug'])
        # Install files for components
        for component in components:
            logger.debug(' # %s', component)
            self.__install(component, context)

    def cookiecutter(self, context, components):
        """Create a new cookiecutter template from the DevOps template given
        config options. Config options only affect the default values for the
        cookiecutter template, which are provided in cookiecutter.json

        Generates components which are defined in template.json

        Params:
            context: Dictionary with configuration flags supported by the
                template (flags are defined in ProjectConfig.cookiecutter and
                can be modified based on command-line args, see
                main.cookiecutter)
            components: Template components that should be installed.
        """
        logger = logging.getLogger('DevOpsTemplate.cookiecutter')
        logger.info('Generate cookiecutter template')
        # Generate cookiecutter.json
        # configuration is provided by the context dictionary
        cookiecutter_json_fpath = os.path.join(self.__projectdir,
                                               'cookiecutter.json')
        try:
            self.__check_project_file(cookiecutter_json_fpath)
            if not self.__dry_run:
                with open(cookiecutter_json_fpath, 'w',
                          encoding='utf-8') as handle:
                    json.dump(context, handle, indent=2)
            logger.info('project:%s', cookiecutter_json_fpath)
        except SkipFileError:
            logger.warning('File %s exists, skipping', cookiecutter_json_fpath)

        # Only generate *cookiecutter* readme if not present already
        # Note: a template readme can be installed via template components
        # (see below)
        readme_fpath = os.path.join(self.__projectdir, 'README.md')
        if not os.path.exists(readme_fpath):
            if not self.__dry_run:
                with open(readme_fpath, 'w', encoding='utf-8') as handle:
                    readme_content_list = ['# Cookiecutter PyDevops']
                    handle.writelines(readme_content_list)
            logger.info('project:%s', readme_fpath)

        # Generate hooks directory with pre/post generation scripts if required
        # (might be useful in future)
        #
        # Generate project template
        cookiecutter_project_dname = '{{cookiecutter.project_name}}/'
        self.__mkdir(cookiecutter_project_dname)
        # Adjust projectdirectory such that __install installs to the correct
        # directory (projectdirectory represents cookiecutter template root)
        cookiecutter_rootdir = self.__projectdir
        self.__projectdir = os.path.join(self.__projectdir,
                                         cookiecutter_project_dname)
        # Generate cookiecutterconfig for rendering cookiecutter template
        # variables
        # pylint: disable=consider-using-f-string
        # simplifies handling of jinja2 template syntax {{ }}
        cookiecutterconfig = {key: '{{cookiecutter.%s}}' % key
                              for key in context.keys()}
        # Install all template components
        for comp in components:
            self.__install(comp, cookiecutterconfig)

        # Revert project directory to cookiecutter root directory
        self.__projectdir = cookiecutter_rootdir

    def manage(self, context, components):
        """Add functionality/components to an existing project that has been
        created from the DevOps template given configuration options.

        Params:
            context: Dictionary with configuration flags supported by the
                template (flags are defined in ProjectConfig.managa and
                can be modified based on command-line args, see
                main.manage)
            components: Template components that should be installed.
        """
        logger = logging.getLogger('DevOpsTemplate.manage')
        # Install files for components
        for component in components:
            logger.debug(' # %s', component)
            self.__install(component, context)

    def __install(self, template_component, context):
        """Copy and render files for a template component
        Components, i.e., file to install, are defined in 'template.json' which
        is represented by __template_dict.

        Params:
            template_component: String specifying the component to install.
            context: Dictionary with the context for rendering Jinja2
                templates.
        """
        file_list = self.__template_dict[template_component]
        for template_fpath in file_list:
            # Render file path (paths can contain template variables)
            project_fpath = Template(template_fpath).render(**context)
            self.__render(template_fpath, project_fpath, context)

    def __mkdir(self, project_dname):
        """Create a directory within the project if not present

        Params:
            project_dname: String specifying the name of the directory
        """
        logger = logging.getLogger('DevOpsTemplate.__mkdir')
        project_dpath = os.path.join(self.__projectdir, project_dname)
        if not os.path.exists(project_dpath):
            logger.info('creating directory: %s', project_dpath)
            if not self.__dry_run:
                os.makedirs(project_dpath)
        else:
            logger.debug('directory %s exists', project_dpath)

    def __render(self, pkg_fname, project_fname, context):
        """Render template to project according to overwrite/skip class members
        The source file will be used as a Jinja2 template and rendered before
        the rendering result will be written to the target file.

        Params:
            pkg_fname: String specifying the file in the distribution package
            project_fname: String specifying the target file in the project
            context: Dictionary with the context for rendering Jinja2 templates
        Raises:
            FileNotFoundError: if pkg_fname is not available
            FileExistsError: if project_fname already exists in the project
                and skip-exists=False, overwrite-exists=False
        """
        logger = logging.getLogger('DevOpsTemplate.__render')
        pkg_fpath = os.path.join(self.__template_dname, pkg_fname)
        if not pkg.exists(pkg_fpath):
            raise FileNotFoundError(f'File {pkg_fpath} not available in '
                                    'distribution package')
        project_fpath = os.path.join(self.__projectdir, project_fname)
        try:
            self.__check_project_file(project_fpath)
        except SkipFileError:
            logger.warning('File %s exists, skipping', project_fpath)
            return
        if not self.__dry_run:
            # Create parent directories if not present
            parent_dname = os.path.dirname(project_fpath)
            if not os.path.exists(parent_dname):
                os.makedirs(parent_dname)
            # Load and instantiate template
            template = self.__env.get_template(pkg_fname)
            with open(project_fpath, 'w', encoding='utf-8') as project_fh:
                template.stream(**context).dump(project_fh)
        logger.info('template:%s  ->  project:%s', pkg_fname, project_fpath)

    def __check_project_file(self, project_fpath):
        """Check whether the given file can be created in the project without
        conflict. A conflict arises if the file exists and should not be
        skipped or overwritten.

        Params:
            project_fpath: String specifying the path to the file in the
                project.
        Returns: True if the file can be created without conflict.
        Raises:
            SkipFileError: if the creation of the new file should be skipped.
            FileExistsError: if the file that should be created already exists
                and should not be overwritten.
        """
        if os.path.exists(project_fpath) and self.__skip:
            raise SkipFileError(f'File {project_fpath} already exists, skip.')
        if os.path.exists(project_fpath) and not self.__overwrite:
            raise FileExistsError(f'File {project_fpath} already exists, exit.'
                                  ' (use --skip-exists or --overwrite-exists'
                                  ' to control behavior)')
        return True



"""Setuptools package configuration file

The 'Makefile' triggers builds and uses setup.py/requirements.txt
Adapt the Makefile variable SETUPTOOLSFILES if dependencies change, e.g.,
you switch to using only setup.py and/or setup.cfg for defining your package.
"""
import os
import re
from setuptools import setup, find_packages


def parse_version(package_rpath):
    """Parse version number from __init__.py in top-level import package

    It is assumed that the version is defined as a string and the '=' sign
    is surrounded by at most one whitespace character to the left and to the
    right.

    Params:
        package_rpath: String with the relative path to the package
            (with respect to setup.py)
    Returns:
        version: String with the version number
    Raises:
        ValueError: if the parser could not match the version definition
    """
    init_fpath = os.path.join(package_rpath, '__init__.py')
    with open(init_fpath, 'r', encoding='utf-8') as handle:
        init_contents = handle.read()
        ver_re = r"^__version__ ?= ?['\"]([^'\"]*)['\"]"
        match = re.search(ver_re, init_contents, re.M)
        if match:
            version = match.group(1)
            return version
        raise ValueError('Could not parse version string')


def read_file(filepath):
    """Read file as text

    Params:
        filename: relative/absolute path to the file
    Returns:
        contents as str
    """
    with open(filepath, 'r', encoding='utf-8') as handle:
        return handle.read()


# Name of the top-level import package (directory)
PKG_NAME = '{{project_slug}}'

# Parse version
VERSION = parse_version(PKG_NAME)

# Read Readme that will be used as long package description
DESCRIPTION_LONG = read_file('README.md')

# Read list of Python package dependencies
INSTALL_REQUIRES = read_file('requirements.txt').splitlines()

# ATTENTION: the name must match the name of the top-level import package
# see Makefile variable PACKAGE.
# Naming the project as the top-level import package is also consistent with
# conventions.
setup(name=PKG_NAME,
      # The version string will be included in your Python package
      # https://setuptools.readthedocs.io/en/latest/setuptools.html#specifying-your-project-s-version
      version=VERSION,
      python_requires='>= 3.8',
      # Define the package sources.
      packages=find_packages(include=[PKG_NAME, f'{PKG_NAME}.*']),
      # Package dependencies
      install_requires=INSTALL_REQUIRES,
      # Defines dev environment containing development dependencies
      # (for linting, testing, etc.)
      extras_require={'dev': ['pip >= 20.1.1',
                              'wheel',
                              'pytest',
                              'pytest-mock',
                              'coverage',
                              'bandit',
                              'pylint',
                              'autopep8',
                              'flake8']
                      },
      # Include scripts/executables for application from 'scripts' directory
      # Executables will be included in the PATH search directory of the Python
      # environment, e.g., a virtual environment or /usr/local/bin
      # scripts=['scripts/samplescript'],
      # Generate entry points (executables) automatically from Python functions
      #     executable = package.module:function
      # Executables will be included in the PATH search directory of the Python
      # environment, e.g., a virtual environment or /usr/local/bin
      entry_points={
          'console_scripts': [
              f'{PKG_NAME}={PKG_NAME}.main:main',
          ],
      },
      # Data files should always be part of the package and you should avoid
      # dependencies to data files outside of the package.
      # find_packages (see above) searches Python packages and includes source
      # files. In order to include data files, too, you can use
      # include_package_data OR package_data
      # Note that package_data will not work when include_package_data=True
      # With include_package_data you must manage your data file includes in
      # MANIFEST.in. With package_data you manage data file includes relative
      # to package key provided in the dictionary.
      # https://setuptools.readthedocs.io/en/latest/setuptools.html#including-data-files
      # include_package_data = True,
      # package_data={
      #     # If any package contains *.txt or *.rst files, include them:
      #     '': ['*.txt', '*.rst'],
      #     # And include any *.msg files found in the 'hello' package, too:
      #     'hello': '*.msg'
      # },
      # Generally do not assume that the package can safely be run as a zip
      # archive
      zip_safe=False,
      # Package meta information
      author='{{author_name}}',
      author_email='{{author_email}}',
      description='{{project_description}}',
      long_description=DESCRIPTION_LONG,
      long_description_content_type='text/markdown',
      # keywords = 'keyword1, keyword2, keyword3'
      url='{{project_url}}',
      # project_urls={
      #     'Bug Tracker': 'https://bugs.example.com/HelloWorld/',
      #     'Documentation': 'https://docs.example.com/HelloWorld/',
      #     'Source Code': 'https://code.example.com/HelloWorld/'
      # },
      platforms='any',
      # license='MIT',
      # Define the intended audience of your Python package
      # For a full list of classifiers see: https://pypi.org/classifiers/
      classifiers=[
          'Intended Audience :: Developers',
          #   'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.8',
      ]
      )



"""Setup.py configuration file
"""
import os
import re
import itertools
import json
from setuptools import setup


def parse_version():
    """Parse version number from __init__.py in top-level import package

    It is assumed that the version is defined as a string and the '=' sign
    is surrounded by at most one whitespace character to the left and to the
    right.

    Returns:
        version string
    Raises:
        ValueError if the parser could not match the version definition
    """
    init_fpath = os.path.join('devopstemplate', '__init__.py')
    with open(init_fpath, 'r', encoding='utf-8') as handle:
        init_contents = handle.read()
        ver_re = r"^__version__ ?= ?['\"]([^'\"]*)['\"]"
        match = re.search(ver_re, init_contents, re.M)
        if match:
            version = match.group(1)
            return version
        raise ValueError('Could not parse version string')


def parse_template_index():
    """Load json that defines template structure
    """
    with open('devopstemplate/template.json', 'r', encoding='utf-8') as handle:
        template_dict = json.load(handle)
        template_index = list(itertools.chain(*template_dict.values()))
    return template_index


def read_file(filepath):
    """Read file as text

    Params:
        filename: relative/absolute path to the file
    Returns:
        contents as str
    """
    with open(filepath, 'r', encoding='utf-8') as handle:
        return handle.read()


# Parse version
VERSION = parse_version()
# Read list of template files that will be packaged for installation with the
# devopstemplate tool
TEMPLATE_INDEX = parse_template_index()

# Read Readme that will be used as long package description
DESCRIPTION_LONG = read_file('README.md')

# Read list of Python package dependencies
INSTALL_REQUIRES = read_file('requirements.txt').splitlines()

# setup.py defines the Python package. The build process is triggered from
# Makefile. Adapt Makefile variable SETUPTOOLSFILES if build file dependencies
# change.
setup(name='devopstemplate',
      version=VERSION,
      python_requires='>= 3.6',
      # Import package
      packages=['devopstemplate'],
      # Package dependencies
      install_requires=INSTALL_REQUIRES,
      # Defines dev environment containing development dependencies
      # (for linting, testing, etc.)
      extras_require={'dev': ['pip >= 20.1.1',
                              'wheel',
                              'pytest',
                              'coverage',
                              'bandit',
                              'pylint',
                              'autopep8',
                              'flake8']
                      },
      entry_points={'console_scripts':
                    ['devopstemplate=devopstemplate.main:main']
                    },
      package_data={
          # Include data files in devopstemplate package
          # the file template.index specifies file paths relative to
          # devopstemplate/template directory
          'devopstemplate': (['template.json', 'commands.json'] +
                             [f'template/{fpath}'
                              for fpath in TEMPLATE_INDEX]),
      },
      # Generally do not assume that the package can safely be run as a zip
      # archive
      zip_safe=False,
      # Metadata to display on PyPI
      author='Leonard Rothacker',
      author_email='leonard.rothacker@googlemail.com',
      description=('This package provides a command-line interface for ' +
                   'setting up a Python project based on a dev-ops template'),
      long_description=DESCRIPTION_LONG,
      long_description_content_type='text/markdown',
      keywords='devops template sonarqube docker code-analysis',
      url='https://github.com/lrothack/dev-ops-admin',
      license='MIT',
      platforms=['any'],
      classifiers=['Intended Audience :: Developers',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   'Programming Language :: Python :: 3.8',
                   'Programming Language :: Python :: 3.9',
                   'Environment :: Console',
                   ]
      )



"""Wrapper around the setuptools pkg_resources package

The module offers convenience function for easily reading strings,
lists of strings, file names and binary stream for files stored in Python
distribution packages.
"""
import pkg_resources


def exists(resource_name):
    """Package resource wrapper for checking if resource exists

    Params:
        resource_name: Relative path to the resource in the package (from the
            package root)
    Returns: boolean specifying existence
    """
    return pkg_resources.resource_exists(__package__,
                                         resource_name)


def isdir(resource_name):
    """Package resource wrapper for checking if resource is a directory

    Params:
        resource_name: Relative path to the resource in the package (from the
            package root)
    Returns: boolean specifying is resource is directory
    """
    return pkg_resources.resource_isdir(__package__,
                                        resource_name)


def filepath(resource_name):
    """Package resource wrapper for obtaining resource filepath

    Attention: resource might have to be copied to a tmp folder in order to
    obtain a path in the file system. Rather work with the resource directly.

    Params:
        resource_name: Relative path to the resource in the package (from the
            package root)
    Returns: absolute path to the resource in the file system
    """
    return pkg_resources.resource_filename(__package__,
                                           resource_name)


def string(resource_name, encoding='utf-8'):
    """Package resource wrapper for obtaining resource contents as string.

    Attention: The originally binary string will be interpreted as a text
    string using the given encoding.

    Params:
        resource_name: Relative path to the resource in the package (from the
            package root).
        encoding: String defining the encoding for interpreting the binary
            string. None refers to the default encoding (typically UTF-8).
    Returns: contents of of resource interpreted as text string (default
        encoding)
    """
    resource_string = pkg_resources.resource_string(__package__,
                                                    resource_name)
    return resource_string.decode(encoding)


def string_list(resource_name, encoding='utf-8'):
    """Package resource wrapper for obtaining resource contents as list of
    strings.

    The function uses the 'string' method and splits the resulting string
    in lines.

    Params:
        resource_name: Relative path to the resource in the package (from the
            package root)
        encoding: String defining the encoding for interpreting the binary
            string. None refers to the default encoding (typically UTF-8).
    Returns: contents of of resource interpreted as list of text strings
        (default encoding)
    """
    return string(resource_name, encoding).splitlines()


def stream(resource_name):
    """Package resource wrapper for obtaining file handle in order to read
    resource contents.

    Params:
        resource_name: Relative path to the resource in the package (from the
            package root)
    Returns: file object for reading resource contents in binary mode
    """
    return pkg_resources.resource_stream(__package__,
                                         resource_name)



import sys
import platform
import logging
import argparse
import {{project_slug}}


def parse_args(args_list):
    """Parse command-line arguments

    Params:
        args_list: list of strings with command-line flags (sys.argv[1:])
    """
    logger = logging.getLogger(f'{__name__}:parse_args')
    description = '{{project_description}}'
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('--quiet', action='store_true',
                        help='Only print warning/error messages')
    parser.add_argument('--verbose', action='store_true',
                        help='Print debug messages')
    #
    # Add more command-line arguments and/or sub-commands
    #

    # Parse arguments in order to obtain results in argparse.Namespace object
    args_ns = parser.parse_args(args_list)

    # Set log level according to command-line flags
    if args_ns.verbose:
        {{project_slug}}.LOGCONFIG.debug()
        logger.debug('%s:: %s\n', platform.node(), ' '.join(sys.argv))
    elif args_ns.quiet:
        {{project_slug}}.LOGCONFIG.warning()

    return args_ns


def main():
    """Entry point for the command-line interface"""
    logger = logging.getLogger(f'{__name__}:main')
    args_ns = parse_args(sys.argv[1:])
    logger.info(args_ns)
    # Start application according to parsing result in args_ns


if __name__ == "__main__":
    main()



"""Defines template creation and modification wrt user configurations.
"""

import os
import subprocess
import shutil
import json
from jinja2 import Template
from devopstemplate import pkg


class CommandsConfig():
    """Represents definitions of optional parameters for (sub-)commands
    create, manage, cookiecutter
    """

    # Static variable storing the command definitions
    __commands_dict = None

    def __init__(self):
        """Loads dictionary with command definitions, if not already present
        """
        if not CommandsConfig.__commands_dict:
            CommandsConfig.__commands_dict = self.__load_commands_dict()
        self.__commands_dict = CommandsConfig.__commands_dict

    def values(self, command, section):
        """Obtain a section for a command from the json configuration

        Params:
            command: String specifying the command (create, manage,
                cookiecutter)
            section: String specifying the configuration section (template,
                parameters, components)
        Returns: List of string xor dict objects containing the
            options/components of the configurations section, empty list if
            the section is not available
        """
        cmd_dict = self.__commands_dict[command]
        try:
            result_list = cmd_dict[section]
        except KeyError:
            result_list = []

        return result_list

    @staticmethod
    def __load_commands_dict():
        """Load commands definition from package resource

        Returns:
            commands_dict: Dictionary representing commands.json
        """
        # Load commands.json with definitions for command-line arguments
        # -> sub-commands, their arguments, defaults and help messages
        commands_fname = 'commands.json'
        commands_str = pkg.string(commands_fname)
        # Define context for substituting default values in the commands file
        # in order to be substituted with Jinja2
        # git user and email as default for author data
        name, email = CommandsConfig.git_user()
        context = {'git_name': name, 'git_email': email}
        commands_str = Template(commands_str).render(**context)
        commands_dict = json.loads(commands_str)
        return commands_dict

    @staticmethod
    def git_user():
        """Obtain git user name and email from git config.

        The function expects that the command 'git' is found on the PATH.

        Returns:
            name: String with git user name, empty string if no 'git' command
            email: String with git user email, empty string if no 'git' command
        """
        name = ''
        email = ''
        if shutil.which('git'):
            name = subprocess.check_output(['git', 'config', 'user.name'],
                                           stderr=subprocess.STDOUT,
                                           encoding='utf-8')
            email = subprocess.check_output(['git', 'config', 'user.email'],
                                            stderr=subprocess.STDOUT,
                                            encoding='utf-8')
        return name.strip(), email.strip()


class ProjectConfig():
    """Generate config dictionaries for template actions:
    create, manage, cookiecutter

    Supports a CLI for asking configuration values from the user in interactive
    mode.
    """

    def __init__(self, args):
        """Initialize config with defaults that all template actions have in
        common

        Params:
            args: argparse.Namespace object with command-line arguments
        """
        # Generate dict from namespace attributes (save dict as a copy)
        self.__args_dict = dict(vars(args))
        self.project_dir = os.path.abspath(args.project_dir)
        self.overwrite_exists = args.overwrite_exists
        self.skip_exists = args.skip_exists
        self.dry_run = args.dry_run
        # load definition for (sub-)commands
        self.__cfg = CommandsConfig()

    def __param_dict(self, command):
        """Generate project configuration for creating an instance of the
        devops template. Defines the context for rendering Jinja2 templates.

        Supports interactive mode.

        Params:
            command: String specifying the command
        Returns:
            param_dict: Dictionary with configurations
        """
        param_def_list = self.__cfg.values(command, 'parameters')
        param_key_list = [p['name'].replace('-', '_') for p in param_def_list]
        # Define param dict for project context
        param_dict = {key: self.__args_dict[key] for key in param_key_list}
        # Overwrite default/specified values in interactive mode
        if ('interactive' in self.__args_dict and
                self.__args_dict['interactive']):
            param_dict = self.__input(param_key_list, param_dict)
        return param_dict

    def __comp_list(self, command):
        """Generate list of project components that will be installed according
        to template.json

        Supports interactive mode.

        Params:
            command: String specifying the command.
        Returns:
            comp_list: List of strings specifying template components
        """
        comp_def_list = self.__cfg.values(command, 'components')
        # Generate internal identifiers (keys) for component by their names
        comp_key_list = [c['name'].replace('-', '_') for c in comp_def_list]
        # comp_dict: dict of project component dicts
        # (keys: name, default, help, template)
        # 'default' defines how to interpret (positive or
        # negative) boolean component flags provided by the user,
        # i.e. for example, add-* or no-*
        # add-*, default: false -> do not add comp. by default, only with flag
        # no-*,  default: true  -> add comp. by default, do not add with flag
        # 'template' is a list of template identifier for installing component
        # files defined in template.json
        comp_dict = dict(zip(comp_key_list, comp_def_list))

        # Define comp_args dict for project component flags specified by user
        comp_args = {key: self.__args_dict[key] for key in comp_key_list}
        # Overwrite default/specified values in interactive mode
        if ('interactive' in self.__args_dict and
                self.__args_dict['interactive']):
            comp_args = self.__input(comp_key_list, comp_args)
        # Translate flags (pos and neg) to positive list of components
        # Initialize with default components for command
        # Create a copy since the list will be extended
        comp_list = list(self.__cfg.values(command, 'template'))
        for key in comp_key_list:
            # Whether to add the component by default (when no user flag has
            # been set)
            comp_default = comp_dict[key]['default']
            # Negate the default boolean value if a flag has been set for the
            # component by the user)
            comp_add = not comp_default if comp_args[key] else comp_default
            if comp_add:
                comp_list.extend(comp_dict[key]['template'])

        return comp_list

    def create(self):
        """Generate project configuration for action 'create'.

        Supports interactive mode.

        Returns:
            param_dict: Dictionary with configurations for creating an instance
                of the devops template.
            comp_list: List of template components that will be installed
                according to template.json
        """
        # Set default for package_name if not present
        project_name = os.path.basename(self.project_dir)
        # If package_name is not given as CLI argument
        if not self.__args_dict['package_name']:
            # Slugify project_name and use as default for package_name
            project_slug = project_name.replace(' ', '_').replace('-', '_')
            project_slug = project_slug.lower()
            self.__args_dict['package_name'] = project_slug

        # Parameters
        param_dict = self.__param_dict(command='create')
        # Set project name
        param_dict['project_name'] = project_name
        # Define project package name: slugify given package_name
        # Default package_name is defined above or given as CLI arg
        # and can be overwritten in interactive mode in __param_dict func
        # --> define project_slug based on package_name in param_dict var
        project_slug = param_dict['package_name']
        project_slug = project_slug.replace(' ', '_').replace('-', '_')
        param_dict['project_slug'] = project_slug.lower()

        # Components
        comp_list = self.__comp_list(command='create')

        return param_dict, comp_list

    def manage(self):
        """Generate project configuration for action 'manage'.

        Returns:
            param_dict: Dictionary with configurations for modifying an
                instance of the devops template.
            comp_list: List of template components that will be added to the
                template instance according to template.json
        """
        # Parameters
        param_dict = {'project_name': os.path.basename(self.project_dir)}

        # Components
        comp_list = self.__comp_list(command='manage')

        return param_dict, comp_list

    def cookiecutter(self):
        """Generate project configuration for action 'cookiecutter'.

        Returns:
            param_dict: Dictionary with configurations for creating a
                cookiecutter template from the devops template.
            comp_list: List of components that will be added to the
                cookiecutter template according to template.json
        """
        # Parameters
        param_dict = self.__param_dict(command='cookiecutter')
        # Set standard cookiecutter definition for project_slug
        project_slug = ''.join(["{{ ",
                                "cookiecutter.project_name.lower()",
                                ".replace(' ', '_')",
                                ".replace('-', '_')",
                                " }}"])
        param_dict['project_slug'] = project_slug

        # Components
        comp_list = self.__comp_list(command='cookiecutter')

        return param_dict, comp_list

    @staticmethod
    def __input(key_list, params):
        """Query the user for a list of configuration options over the
        command-line. Supports boolean options: User input [y]es and [n]o.
        Default values can be confirmed with 'enter'.

        Params:
            key_list: List of strings with configurations options/keys.
            params: Dictionary the contains mappings from configuration options
                to default values.
        Returns:
            params: Dictionary the contains mappings from configuration options
                to values that have be update/confirmed by the user.

        """
        for key in key_list:
            value = params[key]
            key_disp = key.replace('_', '-')
            if isinstance(value, bool):
                # Convert boolean value to string
                value = 'y' if value else 'n'
                value_user = input(f'{key_disp} [ {value} ] : ')
                # Convert string value_user to boolean
                value_user = value_user == 'y'
            else:
                value_user = input(f'{key_disp} [ "{value}" ] : ')
                # Keep initial value if user just presses enter
                # use the user input otherwise
                value_user = value if value_user == '' else value_user
            params[key] = value_user
        return params
